unit comacode9;
               //(for AA version 3.0 or later)
interface

uses windows, sysutils, math, dialogs, forms;

function pi_initialize(funcaddress:pointer; menuname,description:pchar;
                      whandle:HWND):integer; stdcall;
procedure pi_activate(pimage:pointer; whandle:HWND); stdcall;
procedure pi_finalize(whandle:HWND); stdcall;
procedure CreateModel;
procedure GetCoordinates;

implementation

uses frmCC3;

const
 PACTIVEIMAGE = 4096;
 PSELECTIMAGE = 4097;
 PEQU_STD = 1000;
 PEQU_GAU = 1001;
 PCGREEN = 0;
 PCRED = 2;
 BCBLUE = 3;
 MAXRADIUSARRAY = 40000;

type

Tcommand = (ac_getbuffer, ac_getsize, ac_redraw, ac_close, ac_savefits,
            ac_copy, ac_writeheader, ac_readheader, ac_prepareundo,
            ac_thresholds, ac_transferfunc, ac_getselpoint, ac_getselpoint2,
            ac_gethwnd, ac_create, ac_open, ac_getimage, ac_getversion,
            ac_rename, ac_getname, ac_getra, ac_getdec, ac_getx, ac_gety,
            ac_modified);

Tfloatarray = array[0..$fffffff] of single;
pfloatarray = ^Tfloatarray;
Tcallbackfunc = function(command:Tcommand; p1,p2,p3:pointer):integer; stdcall;

(* Commands: see docs

NAME             PARAMS TYPE                   RESULT

ac_getbuffer    (pimage, &buffer, colorplane)  0,2
ac_getsize      (pimage, &x, &y)               0,1
ac_redraw       (pimage, calcstats, autovis)   0,1
ac_close        (pimage, askuser, nil)         0,1
ac_savefits     (pimage, pfilename, nil)       0,1
ac_copy         (pimage, nil, nil)             0,1
ac_writeheader  (pimage, pheader, nil)         0,1
ac_readheader   (pimage, nil, nil)             pheader
ac_prepareundo  (pimage, nil, nil)             0,1
ac_thresholds   (pimage, &min, &max)           0,1
ac_transferfunc (pimage, trf, nil)             0,1
ac_getselpoint  (pimage, &x, &y)               0,1
ac_getselpoint2 (pimage, &x, &y)               0,1
ac_gethwnd      (pimage, &HWND, nil)           0,1
ac_create       (x,y,colored)                  pimage
ac_open         (pfilename, nil, nil)          pimage
ac_getimage     (selection, nil, nil)          pimage

ac_getversion   (nil, nil, nil)                300+
ac_rename       (pimage, name, path)           0,1
ac_getname      (pimage, path, nil)            pfilename
ac_getra        (pimage, x*100, y*100)         ra*1e5
ac_getdec       (pimage, x*100, y*100)         dec*1e5
ac_getx         (pimage, ra*1e6, dec*1e6)      x*100
ac_gety         (pimage, ra*1e6, dec*1e6)      y*100
ac_modified     (pimage, modified, nil)        0,1
*)


var
 callback: Tcallbackfunc;
 newbuffer,oldbuffer: pfloatarray;
 peak: single;
 dimx,dimy: integer;

(*  MAIN VARIABLES DESCRIPTION
 -----------------------------------------------------------------
 dimx, dimy = x and y resolution of the image
 xz, yz     = central pixel coma coordinates
 maxrad     = maximum radius to calculate the model
 peak       = ADU value at the pixel (xz, yz)
 sigma      = sigma value for the Gauss filter
 *)

(*
-------------------------------------------------------------------------------
Functions exported by the DLL
-------------------------------------------------------------------------------
*)

function pi_initialize(funcaddress:pointer; menuname,description:pchar;
         whandle: HWND):integer; stdcall;
begin
  callback := funcaddress;
  strcopy(menuname,'MedComet Coma Model');
  strcopy(description,'Syntetize a comet coma model from a real image');
  Form3 := nil;
  Application.Handle := GetWindow(whandle,GW_OWNER);
  result := 1;
end;

procedure pi_finalize(whandle:HWND); stdcall;
begin
   if Form3 <> nil then Form3.Release;
end;

procedure pi_activate(pimage:pointer; whandle:HWND); stdcall;
begin
  if Form3 = nil then begin
     Form3 := TForm3.Create(Application);
  end;
  Form3.Show;
  GetCoordinates;
end;

// ---------------------------

procedure GetCoordinates;
var
 x,y: integer;
 pimage: pointer;
begin
  pimage := pointer(callback(ac_getimage,pointer(PACTIVEIMAGE),nil,nil));
  x := 0; y := 0;
  if pimage <> nil then callback(ac_getselpoint,pimage,@x,@y);
  Form3.EditX.Text := IntToStr(x);
  Form3.EditY.Text := IntToStr(y);
end;

function getPixel(x,y: integer): single;
begin
   if x < 0 then x := 0;
   if y < 0 then y := 0;
   if x >= dimx then x := dimx-1;
   if y >= dimy then y := dimy-1;
   result := oldbuffer[x+y*dimx];
end;

procedure setPixel(x,y: integer; v:single);
begin
   if (x < 0) or (y < 0) or (x >= dimx) or (y >= dimy) then exit;
   newbuffer[x+y*dimx] := v;
end;

procedure WarningDlg(s: string);
begin
  MessageBox(Application.Handle,pchar(s),'Warning',
             MB_ICONWARNING or MB_OK or MB_TOPMOST or MB_TASKMODAL);
end;

procedure ErrorDlg(s: string);
begin
  MessageBox(Application.Handle,pchar(s),'Error',
             MB_ICONERROR or MB_OK or MB_TOPMOST or MB_TASKMODAL);
end;

procedure InfoDlg(s: string);
begin
  MessageBox(Application.Handle,pchar(s),'Information',
             MB_ICONINFORMATION or MB_OK or MB_TOPMOST or MB_TASKMODAL);
end;

procedure QuickSort(var a: array of single; iLo,iHi: Integer);
var
 Mid,T: single;
 Lo,Hi: integer;
begin
  Lo := iLo;
  Hi := iHi;
  Mid := a[(Lo + Hi) shr 1];
  repeat
     while a[Lo] < Mid do Inc(Lo);
     while a[Hi] > Mid do Dec(Hi);
     if Lo <= Hi then begin
        T := a[Lo];  a[Lo] := a[Hi]; a[Hi] := T;
        Inc(Lo);
        Dec(Hi);
     end;
  until Lo > Hi;
  if Hi > iLo then QuickSort(a, iLo, Hi);
  if Lo < iHi then QuickSort(a, Lo, iHi);
end;

(*
-------------------
FUNCTION GetMedian
-------------------   Returns the Median for a Sorted Array of Single.                                                               *
*)

function getMedian (SortedX: array of Single; N:integer): Single;
begin
  if N <= 0 then
    result := 0
  else if N = 1 then // Only a Single Value
    Result := SortedX [0]
  else if Odd (N) then // Handle Odd Number of Values
    Result := SortedX [N div 2]
  else // Handle Even Number of Values
    Result := (SortedX [N div 2 - 1] + SortedX [N div 2]) / 2;
end;


(*
-------------------
FUNCTION medCircle
-------------------
Returns the median value of a circle of radius = radius
and center (xz,yz)
*)

function medCircle(xz,yz,radius: integer): single;
var
 x,y,count: integer;
 circArray: array[0..MAXRADIUSARRAY] of Single;
begin
  count := 0;
  for y := yz-radius to yz+radius do
    for x := xz-radius to xz+radius do begin
        if (x < 0) or (y < 0) or (x >= dimx) or (y >= dimy) then continue;
        if round(sqrt((x-xz)*(x-xz)+(y-yz)*(y-yz))) = radius then begin
           circArray[count] := getPixel(x,y);
           count := count + 1;                   // Populating the array
           if count >= High(circArray) then begin
              result := 0;
              exit;
           end;
        end;
    end;
   QuickSort(circArray,0,count-1);
   Result := getMedian(circArray,count);         // Calculating the median value
end;

// -----------------------------------------------

function Interpolate(const arr: array of single; x:single): single;
var
 i: integer;
 c: single;
begin
  i := round(x-0.5);
  c := x - i;
  result := arr[i]*(1-c) + arr[i+1]*c;
end;

(*
---------------------
PROCEDURE createmodel
---------------------    Create the image model of the coma
*)

procedure CreateModel;
var
 pnewimage,pimage: pointer;
 x,y,b,radius: integer;
 xz,yz,maxrad: integer;
 r: single;
 circles: array[0..MAXRADIUSARRAY] of single;
begin
  pimage := pointer(callback(ac_getimage,pointer(PACTIVEIMAGE),nil,nil));
  if pimage = nil then begin
     ErrorDlg('Please open an image.');   // no images opened
     exit;
  end;

  with Form3 do begin
    peak := StrToInt(EditPeak.Text);
    xz := StrToInt(EditX.Text);
    yz := StrToInt(EditY.Text);
    maxrad := StrToInt(EditMaxRad.Text);
  end;
  if (xz = 0) and (yz = 0) then begin
     InfoDlg('Please select a point as coma centroid');
     exit;
  end;

  callback(ac_getsize,pimage,@dimx,@dimy);   // get image size
  callback(ac_prepareundo,pimage,nil,nil);
  getmem(oldbuffer,dimx*dimy*4);

  b := callback(ac_getbuffer,pimage,@newbuffer,pointer(PCGREEN));  //***
  if b <> 2 then exit;     // buffer is not 32 bit, old version of Astroart
  copymemory(oldbuffer,newbuffer,dimx*dimy*4); // copy image to  an another buffer

  pnewimage := pointer(callback(ac_create,pointer(dimx),pointer(dimy),nil));
  b := callback(ac_getbuffer,pnewimage,@newbuffer,pointer(PCGREEN));
  if b <> 2 then exit;

  // Creation of the circular coma model with radius = maxrad
  for radius := 1 to maxrad do begin
      circles[radius] := medCircle(xz,yz,radius);
  end;
  circles[0] := circles[1];
  circles[maxrad+1] := circles[maxrad];

  (*
  for radius := 1 to maxrad do begin
      for y := yz-radius to yz+radius do     // versione senza interpolazione
      for x := xz-radius to xz+radius do begin
          if (x < 0) or (y < 0) or (x >= dimx) or (y >= dimy) then continue;
          if round(sqrt((x-xz)*(x-xz)+(y-yz)*(y-yz))) = radius
             then setPixel(x,y,circles[radius]);
      end;
  end;   *)

  for y := yz-maxrad to yz+maxrad do
    for x := xz-maxrad to xz+maxrad do begin
        if (x < 0) or (y < 0) or (x >= dimx) or (y >= dimy) then continue;
        r := sqrt((x-xz)*(x-xz)+(y-yz)*(y-yz));
        if r >= maxrad then continue;
        setPixel(x,y,Interpolate(circles,r));
    end;

  setPixel(xz,yz,peak);  //Complete the coma model with the
                          //original central peak value
  freemem(oldbuffer);
  callback(ac_redraw,pnewimage,pointer(1),pointer(1));  // update the filtered image
end;

end.

